/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/


#ifndef _QUNITP_H_
#define _QUNITP_H_

#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <memory.h>

/** Some useful constants  **/

#define QUNIT_TRUE            1
#define QUNIT_FALSE           0

#define QUNIT_CATALOGUE_ENTRIES  28   /* size of unit description table */


/** Global var for level of error messages **/

enum QunitVerbosity {
	QUNIT_VERBOSITY_DEBUG,      /* chat on stdout, plus below */
	QUNIT_VERBOSITY_WARNINGS,   /* warnings to stderr, plus below */
	QUNIT_VERBOSITY_QUIET       /* errors to stderr */
};

extern enum QunitVerbosity QUNIT_VERBOSITY;




/**  Unit description structure for catalogue of recognized units  **/

typedef struct QunitUnitDescription {

	const char* symbol;
	const char*   name;
                        /* indices to express unit in terms of */
        int    m;       /* the seven official SI base units... */
        int   kg;
        int    s;
        int    A;
        int    K;
        int  mol;
        int   cd;
        int  rad;       /* ... and the two SI supplementary units */
        int   sr;
        int  deg;       /* this is a non-standard extension we want */

} qunit_unit_description_t;



/** Symbol table element to hold unit expressions and constants **/


typedef struct QunitUnitEntry {

    struct QunitUnitEntry       *next;      /*  points to next entry        */
    struct QunitUnitDescription *SI_unit;   /*  points to catalogue entry   */
    float  exponent;                        /*  allow non-integer exponents */
    char*  text;                            /*  for non-catalogued items    */

} qunit_unit_entry_t;



/** Structure to describe a parsed SI_conversion attribute **/

typedef struct QunitSIConversion {

        float  value;
        struct QunitUnitEntry *first_entry;

        float    m;       /* cumulative base-unit-eqvt exponents */
        float   kg;
        float    s;
        float    A;
        float    K;
        float  mol;
        float   cd;
        float  rad;
        float   sr;
        float  deg;   

}  qunit_si_conversion;


/**  qunit catalogue routines  **/

void QunitInitialiseCatalogue ( );
void QunitPrintCatalogueDescription (const struct QunitUnitDescription*);
struct QunitUnitDescription* QunitLookupCatalogueDescription (const char*);



/**  qunit structure creation, destruction and output routines  **/

struct QunitSIConversion* QunitCreateSIConvDescrip ( );
void QunitDestroySIConvDescrip (struct QunitSIConversion*);

struct QunitUnitEntry*    QunitCreateUnitEntry ( );
void  QunitDestroyUnitEntry (struct QunitUnitEntry*);

void QunitStrCatSIConversionBase (char*, struct QunitSIConversion*);
void QunitStrCatSIConversionTidy (char*, struct QunitSIConversion*);
void QunitStrCatUnitEntry    (char*, struct QunitUnitEntry*);
void QunitStrCatPowerTidy (char* str_buf, struct QunitSIConversion *SIc, double p);


/**  qunit structure accumulating routines  **/

struct QunitUnitEntry* QunitAccumulateUnitEntry
                          (struct QunitUnitEntry*, struct QunitUnitEntry*);

struct QunitSIConversion* QunitAccumulateSIConversion
                                               (struct QunitSIConversion*);


/**  parsing routine  **/

int  QunitSIConvParse (struct QunitSIConversion*);

char * qunit_findNext();

void QunitAddUnit(struct QunitSIConversion * SI_c, 
                  char * unit, 
		  float power);
		  
double Qar_pow(double x, double y);

#endif
